/*
 * et-h-gssapi_err_generic.h:
 * This file is automatically generated; please do not edit it.
 */

#include <et/com_err.h>

#define G_BAD_SERVICE_NAME                       (-2045022976L)
#define G_BAD_STRING_UID                         (-2045022975L)
#define G_NOUSER                                 (-2045022974L)
#define G_VALIDATE_FAILED                        (-2045022973L)
#define G_BUFFER_ALLOC                           (-2045022972L)
#define G_BAD_MSG_CTX                            (-2045022971L)
#define G_WRONG_SIZE                             (-2045022970L)
#define G_BAD_USAGE                              (-2045022969L)
#define G_UNKNOWN_QOP                            (-2045022968L)
#define G_NO_HOSTNAME                            (-2045022967L)
#define G_BAD_HOSTNAME                           (-2045022966L)
#define G_WRONG_MECH                             (-2045022965L)
#define G_BAD_TOK_HEADER                         (-2045022964L)
#define G_BAD_DIRECTION                          (-2045022963L)
#define G_TOK_TRUNC                              (-2045022962L)
#define G_REFLECT                                (-2045022961L)
#define G_WRONG_TOKID                            (-2045022960L)
#define G_CRED_USAGE_MISMATCH                    (-2045022959L)
#define G_STORE_ACCEPTOR_CRED_NOSUPP             (-2045022958L)
#define G_STORE_NON_DEFAULT_CRED_NOSUPP          (-2045022957L)
extern const struct error_table et_ggss_error_table;
extern void initialize_ggss_error_table(void);

/* For compatibility with Heimdal */
extern void initialize_ggss_error_table_r(struct et_list **list);

#define ERROR_TABLE_BASE_ggss (-2045022976L)

/* for compatibility with older versions... */
#define init_ggss_err_tbl initialize_ggss_error_table
#define ggss_err_base ERROR_TABLE_BASE_ggss
